#include <linux/module.h> 
#include <linux/device.h>
#include <linux/cdev.h>

#define MODULE_NAME "sample_driver"
#define DEVICE_NAME "devsample"
#define CLASS_NAME "sample"
#define DEV_FILE "samplehw"

#define SAMPLE_MINOR_BASE 0
#define SAMPLE_MINOR_COUNT 3

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("This is a sample driver.");
MODULE_AUTHOR("Yutaka Hirata");

struct sample_driver {
	struct device_driver driver;
};

static int g_major_num = 0;

static struct class *my_class;
static struct cdev my_cdev;

static char *sample_devnode(struct device *dev, umode_t *mode)
{
	int minor = MINOR(dev->devt);

	printk("%s: %p (%d)\n", __func__, mode, minor);

	if (mode && minor == 0) {
		*mode = 0666;
		printk("mode %o\n", *mode);
	}
	return NULL;
}

static int sample_open(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);

	return 0;
}

static int sample_close(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);

	return 0;
}

static const struct file_operations sample_fops = {
    .open = sample_open,
    .release = sample_close,
};

static int sample_init(struct sample_driver *drv)
{
	int ret;
	int registered = 0;
	int added = 0;
	int classed = 0;
	int i;
	int minor_count = 0;
	dev_t devid;
	struct device *my_device;

	printk("driver loaded\n");

	ret = alloc_chrdev_region(&devid, SAMPLE_MINOR_BASE, SAMPLE_MINOR_COUNT, DEVICE_NAME);
	if (ret < 0) {
		printk("alloc_chrdev_region error %d(%pe)\n", ret, ERR_PTR(ret));
		goto error;
	}
	g_major_num = MAJOR(devid);
	printk("%s: Major number %d was assigned\n", __func__, g_major_num);
	registered = 1;

	cdev_init(&my_cdev, &sample_fops);
	ret = cdev_add(&my_cdev, devid, SAMPLE_MINOR_COUNT);
	if (ret < 0) {
		printk("cdev_add error %d(%pe)\n", ret, ERR_PTR(ret));
		goto error;
	}
	added = 1;

	my_class = class_create(THIS_MODULE, CLASS_NAME);
	if (IS_ERR(my_class)) {
		printk("register_chrdev error (%pe)\n", ERR_CAST(my_class));
		ret = PTR_ERR(my_class);
		goto error;
	}
	my_class->devnode = sample_devnode;
	classed = 1;

	minor_count = 0;
	for (i = 0 ; i < SAMPLE_MINOR_COUNT; i++) {
		my_device = device_create(
						my_class, NULL, 
						MKDEV(g_major_num, SAMPLE_MINOR_BASE + i),
						NULL,
						DEV_FILE"%d", i 
				);
		if (IS_ERR(my_device)) {
			printk("device_create error (%pe)\n", ERR_CAST(my_device));
			ret = PTR_ERR(my_device);
			goto error;
		}
		minor_count++;
	}

	return 0;

error:
	for (i = 0 ; i < minor_count ; i++) {
		device_destroy(my_class, MKDEV(g_major_num, SAMPLE_MINOR_BASE + i));
		printk("%s: device_destroy\n", __func__);
	}

	if (classed) {
		class_destroy(my_class);
		printk("%s: class_destroy\n", __func__);
	}

	if (added) {
		cdev_del(&my_cdev);
		printk("%s: cdev_del\n", __func__);
	}

	if (registered) {
		dev_t dev = MKDEV(g_major_num, SAMPLE_MINOR_BASE);
		unregister_chrdev_region(dev, SAMPLE_MINOR_COUNT);
		printk("%s: unregister_chrdev_region\n", __func__);
	}

	return (ret);
}

static void sample_exit(struct sample_driver *drv)
{
	int i;
	dev_t dev = MKDEV(g_major_num, SAMPLE_MINOR_BASE);

	printk("driver unloaded\n");

	for (i = 0 ; i < SAMPLE_MINOR_COUNT;  i++) {
		device_destroy(my_class, MKDEV(g_major_num, SAMPLE_MINOR_BASE + i));
	}

	class_destroy(my_class);
	cdev_del(&my_cdev);
	unregister_chrdev_region(dev, SAMPLE_MINOR_COUNT);
}

static struct sample_driver sa_drv = {
	.driver = {
		.name = MODULE_NAME,
	},
};

module_driver(sa_drv, sample_init, sample_exit);

