#include <linux/module.h>
#include <linux/device.h>

#define MODULE_NAME "sample_driver"
#define DEVICE_NAME "devsample"
#define CLASS_NAME "sample"
#define DEV_FILE "samplehw"

#define SAMPLE_MAJOR 300
#define SAMPLE_MINOR 0

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("This is a sample driver.");
MODULE_AUTHOR("Yutaka Hirata");

struct sample_driver {
	struct device_driver driver;
};

static int g_major_num = SAMPLE_MAJOR;
module_param(g_major_num, int, S_IRUGO);
MODULE_PARM_DESC(g_major_num, "Major number (default: 300)");

static struct class *my_class;
static struct device *my_device;

/*
 * Sample data
 */
#define BUFFER_MAX 32
static char g_buffer[BUFFER_MAX];
static size_t g_buffer_count;

wait_queue_head_t g_read_wait;

/*
 * TASK_UNINTERRUPTIBLE sleepにする場合は下記マクロを定義する。
 */
//#define SLEEP_D_STAT

/*
 * mutext
 */
static DEFINE_MUTEX(g_mutex);

static inline bool buffer_readable(void)
{
	return (g_buffer_count > 0);
}

static char *sample_devnode(struct device *dev, umode_t *mode)
{
	int minor = MINOR(dev->devt);

	printk("%s: %p (%d)\n", __func__, mode, minor);

	if (mode && minor == 0) {
		*mode = 0666;
		printk("mode %o\n", *mode);
	}
	return NULL;
}

static int sample_open(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);


	return 0;
}

static ssize_t sample_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
	size_t len;
	unsigned long rem;
	ssize_t written;

	mutex_lock(&g_mutex);

	printk("%s entered.\n", __func__);

	/* 書き込める残バイト数 */
	len = min(BUFFER_MAX - g_buffer_count, count);

	if (len == 0) {
		printk("%s: %zu + %zu overflow\n", __func__, g_buffer_count, count);
		mutex_unlock(&g_mutex);
		return -ENOSPC;
	}

	/*
	 * ユーザー空間からカーネル空間へコピーする。
	 * 返り値が 0 なら成功。
	 * 0 より大きな値なら、コピーできなかったバイト数を表す。
	 *
	 */
	rem = copy_from_user(&g_buffer[g_buffer_count], buf, len);
	printk("copy_from_user %lu\n", rem);
	if (rem) {
		written = (len - rem);
	} else {
		written = len;
	}
	g_buffer_count += written;

	print_hex_dump(KERN_ALERT, "", 0, 16, 1, g_buffer, g_buffer_count, true);

#ifdef SLEEP_D_STAT
	wake_up(&g_read_wait);
#else
	wake_up_interruptible(&g_read_wait);
#endif

	mutex_unlock(&g_mutex);

	return written;
}

static ssize_t sample_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	unsigned long rem;
	size_t copied, res;
	int ret;


	printk("%s entered.\n", __func__);

#ifdef SLEEP_D_STAT
	wait_event(g_read_wait, buffer_readable());
	ret = 0;
#else
	ret = wait_event_interruptible(g_read_wait, buffer_readable());
	if (ret < 0) {
		return -ERESTARTSYS;
	}
#endif

	mutex_lock(&g_mutex);

	copied = min(g_buffer_count, count);
	rem = copy_to_user(buf, g_buffer, copied);
	printk("copy_to_user %lu (copied %zu) (bufsize %zu)\n", rem, copied, g_buffer_count);
	/* 
	 * カーネル空間からユーザー空間へコピーが失敗したらエラーとする。
	 */
	if (rem) {
		mutex_unlock(&g_mutex);
		return -EFAULT;
	}

	/* 
	 * まだバッファが残っている場合は、前方に移動させる。
	 */
	res = g_buffer_count - copied;
	if (res > 0) {
		memmove(&g_buffer[0], &g_buffer[copied], res);
	}

	g_buffer_count = res;

	mutex_unlock(&g_mutex);

	/* コピーしたサイズを返す */
    return copied;
}

static int sample_close(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);

	return 0;
}

static const struct file_operations sample_fops = {
    .open = sample_open,
	.write = sample_write,
	.read = sample_read,
    .release = sample_close,
};

static int sample_init(struct sample_driver *drv)
{
	int ret;
	int registered = 0;
	int classed = 0;
	int deviced = 0;

	printk("driver loaded\n");

	ret = register_chrdev(g_major_num, DEVICE_NAME, &sample_fops);
	if (ret) {
		printk("register_chrdev error %d(%pe)\n", ret, ERR_PTR(ret));
		goto error;
	}
	registered = 1;

	my_class = class_create(THIS_MODULE, CLASS_NAME);
	if (IS_ERR(my_class)) {
		printk("register_chrdev error (%pe)\n", ERR_CAST(my_class));
		ret = PTR_ERR(my_class);
		goto error;
	}
	my_class->devnode = sample_devnode;
	classed = 1;

	my_device = device_create(
					my_class, NULL, 
					MKDEV(g_major_num, SAMPLE_MINOR),
					NULL,
					DEV_FILE
			);
	if (IS_ERR(my_device)) {
		printk("device_create error (%pe)\n", ERR_CAST(my_device));
		ret = PTR_ERR(my_device);
		goto error;
	}
	deviced = 1;

	/* waitキューの初期化 */
	init_waitqueue_head(&g_read_wait);
	/* バッファの初期化 */
	g_buffer_count = 0;

	return 0;

error:
	if (deviced) {
		device_destroy(my_class, MKDEV(SAMPLE_MAJOR, SAMPLE_MINOR));
	}

	if (classed) {
		class_destroy(my_class);
	}

	if (registered) {
		unregister_chrdev(g_major_num, DEVICE_NAME);
	}

	return (ret);
}

static void sample_exit(struct sample_driver *drv)
{
	printk("driver unloaded\n");

	device_destroy(my_class, MKDEV(g_major_num, SAMPLE_MINOR));
	class_destroy(my_class);
	unregister_chrdev(g_major_num, DEVICE_NAME);
}

static struct sample_driver sa_drv = {
	.driver = {
		.name = MODULE_NAME,
	},
};

module_driver(sa_drv, sample_init, sample_exit);

