#include <linux/module.h>
#include <linux/device.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/io.h>
#include <linux/interrupt.h>

#define MODULE_NAME "sample_driver"
#define DEVICE_NAME "devsample"
#define CLASS_NAME "sample"
#define DEV_FILE "samplehw"

#define SAMPLE_MAJOR 300
#define SAMPLE_MINOR 0

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("This is a sample driver.");
MODULE_AUTHOR("Yutaka Hirata");

struct sample_driver {
	struct device_driver driver;
};

static int g_major_num = SAMPLE_MAJOR;
module_param(g_major_num, int, S_IRUGO);
MODULE_PARM_DESC(g_major_num, "Major number (default: 300)");

static struct class *my_class;
static struct device *my_device;

/*
 * mutext
 */
static DEFINE_MUTEX(g_mutex);

static void *my_intterupt;
static int my_irq_num = 47;

static DEFINE_SPINLOCK(g_lock);


static void do_critical_section(void)
{
	printk("%s entered.\n", __func__);
}

static char *sample_devnode(struct device *dev, umode_t *mode)
{
	int minor = MINOR(dev->devt);

	printk("%s: %p (%d)\n", __func__, mode, minor);

	if (mode && minor == 0) {
		*mode = 0666;
		printk("mode %o\n", *mode);
	}
	return NULL;
}

static int sample_open(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);

	return 0;
}

static ssize_t sample_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
	printk("%s entered.\n", __func__);

	mutex_lock(&g_mutex); 

	spin_lock_irq(&g_lock);
	do_critical_section();
	spin_unlock_irq(&g_lock);

	mutex_unlock(&g_mutex);

	return count;
}

static ssize_t sample_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	printk("%s entered.\n", __func__);

	mutex_lock(&g_mutex);

	mutex_unlock(&g_mutex);

    return count;
}

static int sample_close(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);

	return 0;
}

static irqreturn_t sample_interrupt(int irq, void *arg)
{
	void *ptr = arg;
	int handled = 0;

	printk("%s entered.\n", __func__);
	printk("ptr %px\n", ptr);

	spin_lock(&g_lock);
	do_critical_section();
	spin_unlock(&g_lock);

	return IRQ_RETVAL(handled);
}

static const struct file_operations sample_fops = {
    .open = sample_open,
	.write = sample_write,
	.read = sample_read,
    .release = sample_close,
};

static int sample_init(struct sample_driver *drv)
{
	int ret;
	int registered = 0;
	int classed = 0;
	int deviced = 0;

	printk("driver loaded\n");

	ret = register_chrdev(g_major_num, DEVICE_NAME, &sample_fops);
	if (ret) {
		printk("register_chrdev error %d(%pe)\n", ret, ERR_PTR(ret));
		goto error;
	}
	registered = 1;

	my_class = class_create(THIS_MODULE, CLASS_NAME);
	if (IS_ERR(my_class)) {
		printk("register_chrdev error (%pe)\n", ERR_CAST(my_class));
		ret = PTR_ERR(my_class);
		goto error;
	}
	my_class->devnode = sample_devnode;
	classed = 1;

	my_device = device_create(
					my_class, NULL, 
					MKDEV(g_major_num, SAMPLE_MINOR),
					NULL,
					DEV_FILE
			);
	if (IS_ERR(my_device)) {
		printk("device_create error (%pe)\n", ERR_CAST(my_device));
		ret = PTR_ERR(my_device);
		goto error;
	}
	deviced = 1;

	/*
	 * 割り込みハンドラを登録する
	 */
	my_intterupt = &my_irq_num;
	ret = request_irq(my_irq_num, sample_interrupt, IRQF_SHARED, "sample-int", my_intterupt);
	if (ret) {
		printk("request_irq error (%pe)\n", ERR_PTR(ret));
		goto error;
	}

	return 0;

error:
	if (deviced) {
		device_destroy(my_class, MKDEV(SAMPLE_MAJOR, SAMPLE_MINOR));
	}

	if (classed) {
		class_destroy(my_class);
	}

	if (registered) {
		unregister_chrdev(g_major_num, DEVICE_NAME);
	}

	return (ret);
}

static void sample_exit(struct sample_driver *drv)
{
	const void *pret;

	printk("driver unloaded\n");

	pret = free_irq(my_irq_num, my_intterupt);
	if (pret) {
		printk("free_irq() %s\n", (char *)pret);
	}

	device_destroy(my_class, MKDEV(g_major_num, SAMPLE_MINOR));
	class_destroy(my_class);
	unregister_chrdev(g_major_num, DEVICE_NAME);
}

static struct sample_driver sa_drv = {
	.driver = {
		.name = MODULE_NAME,
	},
};

module_driver(sa_drv, sample_init, sample_exit);

