#include <linux/module.h>
#include <linux/device.h>
#include <linux/kthread.h>
#include <linux/delay.h>

#define MODULE_NAME "sample_driver2"
#define DEVICE_NAME "devsample2"
#define CLASS_NAME "sample2"
#define DEV_FILE "samplehw2"

#define SAMPLE_MAJOR 300+2
#define SAMPLE_MINOR 0

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("This is a sample driver.");
MODULE_AUTHOR("Yutaka Hirata");

struct sample_driver {
	struct device_driver driver;
};

static int g_major_num = SAMPLE_MAJOR;
module_param(g_major_num, int, S_IRUGO);
MODULE_PARM_DESC(g_major_num, "Major number (default: 302)");

static struct class *my_class;
static struct device *my_device;

/*
 * mutext
 */
static DEFINE_MUTEX(g_mutex);

/*
 * Kernel thread
 */
static struct task_struct *thread;
static struct kthread_worker kworker;

struct sample_work_data {
	struct kthread_work do_work;
	void *ptr;
	size_t count;
};

static struct sample_work_data workdata;

static char *sample_devnode(struct device *dev, umode_t *mode)
{
	int minor = MINOR(dev->devt);

	printk("%s: %p (%d)\n", __func__, mode, minor);

	if (mode && minor == 0) {
		*mode = 0666;
		printk("mode %o\n", *mode);
	}
	return NULL;
}

static int sample_open(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);


	return 0;
}

static ssize_t sample_write(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
	printk("%s entered.\n", __func__);

	mutex_lock(&g_mutex);

	workdata.ptr = (void *)buf;
	workdata.count = count; 
	kthread_queue_work(&kworker, &workdata.do_work);

	mutex_unlock(&g_mutex);

	return count;
}

static ssize_t sample_read(struct file *file, char __user *buf, size_t count, loff_t *ppos)
{
	printk("%s entered.\n", __func__);

	mutex_lock(&g_mutex);

	mutex_unlock(&g_mutex);

    return count;
}

static int sample_close(struct inode *inode, struct file *filp)
{
	printk("%s entered.\n", __func__);

	return 0;
}

static const struct file_operations sample_fops = {
    .open = sample_open,
	.write = sample_write,
	.read = sample_read,
    .release = sample_close,
};

static void sample_do_proc(struct kthread_work *ws)
{
	struct sample_work_data *wp;

	wp = container_of(ws, struct sample_work_data, do_work);

	printk("%s: %px %zu\n", __func__,
			wp->ptr,
			wp->count
			);

	msleep_interruptible(10 * 1000);

	printk("%s leaved.\n", __func__);
}

static int sample2_init(struct sample_driver *drv)
{
	int ret;
	int registered = 0;
	int classed = 0;
	int deviced = 0;
	int err = 0;

	printk("driver loaded\n");

	ret = register_chrdev(g_major_num, DEVICE_NAME, &sample_fops);
	if (ret) {
		printk("register_chrdev error %d(%pe)\n", ret, ERR_PTR(ret));
		goto error;
	}
	registered = 1;

	my_class = class_create(THIS_MODULE, CLASS_NAME);
	if (IS_ERR(my_class)) {
		printk("register_chrdev error (%pe)\n", ERR_CAST(my_class));
		ret = PTR_ERR(my_class);
		goto error;
	}
	my_class->devnode = sample_devnode;
	classed = 1;

	my_device = device_create(
					my_class, NULL, 
					MKDEV(g_major_num, SAMPLE_MINOR),
					NULL,
					DEV_FILE
			);
	if (IS_ERR(my_device)) {
		printk("device_create error (%pe)\n", ERR_CAST(my_device));
		ret = PTR_ERR(my_device);
		goto error;
	}
	deviced = 1;

	/* カーネルスレッドの起動 */
	kthread_init_worker(&kworker);
	thread = kthread_run(kthread_worker_fn, &kworker, "samplethread2");
	if (IS_ERR(thread)) { /* 起動失敗 */
		err = PTR_ERR(thread);
		thread = NULL;

		printk("unable to create kernel thread %i\n", err);
		ret = err;
		goto error;
	}

	kthread_init_work(&workdata.do_work, sample_do_proc);

	return 0;

error:
	if (deviced) {
		device_destroy(my_class, MKDEV(SAMPLE_MAJOR, SAMPLE_MINOR));
	}

	if (classed) {
		class_destroy(my_class);
	}

	if (registered) {
		unregister_chrdev(g_major_num, DEVICE_NAME);
	}

	return (ret);
}

static void sample2_exit(struct sample_driver *drv)
{
	/* カーネルスレッドを停止する */
	if (thread) {
		kthread_flush_worker(&kworker);
		kthread_stop(thread);
	}

	device_destroy(my_class, MKDEV(g_major_num, SAMPLE_MINOR));
	class_destroy(my_class);
	unregister_chrdev(g_major_num, DEVICE_NAME);

	printk("driver unloaded\n");
}

static struct sample_driver sa_drv = {
	.driver = {
		.name = MODULE_NAME,
	},
};

module_driver(sa_drv, sample2_init, sample2_exit);

